#ifndef  _QV_INPUT_
#define  _QV_INPUT_

#include <QvDict.h>
#include <QvString.h>

class QvNode;
class QvDB;

class __declspec(novtable) QvInput {
 public:

    QvInput();
    ~QvInput();

    virtual int         get() = 0;  //KTG
    virtual void        unget(char c) = 0;  //KTG
    virtual QvBool      eof() const = 0;  //KTG

    static float        isASCIIHeader(const char *string);
    void                reset();
    float               getVersion();
    QvBool              get(char &c);
    QvBool              read(char           &c);
    QvBool              read(QvString       &s);
    QvBool              read(QvName         &n, QvBool validIdent = FALSE);
    QvBool              read(int            &i);
    QvBool              read(unsigned int   &i);
    QvBool              read(short          &s);
    QvBool              read(unsigned short &s);
    QvBool              read(long           &l);
    QvBool              read(unsigned long  &l);
    QvBool              read(float          &f);
    QvBool              read(double         &d);
    void                getLocationString(QvString &string) const;
    void                putBack(char c);
    void                putBack(const char *string);
    void                addReference(const QvName &name, QvNode *node);
    QvNode *            findReference(const QvName &name) const;

  private:
    int                 lineNum;        // Number of line currently reading
    float               version;        // Version number of file
    QvBool              readHeader;     // TRUE if header was checked for A/B
    QvBool              headerOk;       // TRUE if header was read ok
    QvDict              refDict;        // Node reference dictionary
    QvString            backBuf;
    int                 backBufIndex;           

    QvBool              checkHeader();

    QvBool              skipWhiteSpace();

    QvBool              readInteger(long &l);
    QvBool              readUnsignedInteger(unsigned long &l);
    QvBool              readReal(double &d);
    QvBool              readUnsignedIntegerString(char *str);
    int                 readDigits(char *string);
    int                 readHexDigits(char *string);
    int                 readChar(char *string, char charToRead);

friend class QvNode;
friend class QvDB;
};

#ifndef _NO_CRT
class QvFileInput : public QvInput
{
  public:
    QvFileInput() { setFilePointer(stdin); }
    
    virtual int         get() { return getc(fp); }
    virtual void        unget(char c) { ungetc(c, fp); }
    virtual QvBool      eof() const { return feof(fp); }

    void                setFilePointer(FILE *newFP)
    { fp = newFP; reset(); }

    FILE *              getCurFile() const { return fp; }
  protected:
    FILE                *fp;            // File pointer
};
#endif

#endif /* _QV_INPUT_ */
