#ifndef  _QV_NODE_
#define  _QV_NODE_

#include <QvString.h>

class QvChildList;
class QvDict;
class QvFieldData;
class QvInput;
class QvNodeList;
class QvState;

enum QvKind {
    QvGroup_node,
    QvSeparator_node,
    QvLOD_node,
    QvSwitch_node,
    QvTransformSeparator_node,

    QvCoordinate3_node,
    QvFontStyle_node,	
    QvMaterial_node,	
    QvMaterialBinding_node,
    QvNormal_node,		
    QvNormalBinding_node,	
    QvShapeHints_node,		
    QvTextureCoordinate2_node,	
    QvTexture2_node,		
    QvTexture2Transform_node,	

    QvDirectionalLight_node,
    QvPointLight_node,
    QvSpotLight_node,

    QvOrthographicCamera_node,
    QvPerspectiveCamera_node,	

    QvTransform_node,	
    QvRotation_node,	
    QvMatrixTransform_node,
    QvTranslation_node,
    QvScale_node,

	QvAsciiText_node,
	QvCone_node,
	QvCube_node,
	QvCylinder_node,
	QvIndexedFaceSet_node,
	QvIndexedLineSet_node,
	QvPointSet_node,
	QvSphere_node,

	QvWWWAnchor_node,
	QvWWWInline_node,
	QvInfo_node,
	QvUnknownNode_node,
};

class QvNode {

  public:
    enum Stage {
	FIRST_INSTANCE,		// First real instance being constructed
	PROTO_INSTANCE,		// Prototype instance being constructed
	OTHER_INSTANCE		// Subsequent instance being constructed
    };

    QvFieldData	*fieldData;
    QvChildList	*children;
    QvBool	isBuiltIn;

    QvName		*objName;
    QvNode();
    virtual ~QvNode();

    // Reference counting:
    long	refCount;
    void	ref() const;		// Adds reference
    void	unref() const;		// Removes reference, deletes if now 0
    void	unrefNoDelete() const;	// Removes reference, never deletes

    const QvName &	getName() const;
    void		setName(const QvName &name);

    static void		init();
    static QvBool	read(QvInput *in, QvNode *&node);

    virtual QvKind      getKind() = 0;
    virtual QvFieldData *getFieldData() = 0;


    virtual void	traverse(QvState *state) = 0;

  protected:
    virtual QvBool	readInstance(QvInput *in);

  private:
    static QvDict	*nameDict;
    QV_FREE_STATIC_MEMBER
    static void		addName(QvNode *, const char *);
    static void		removeName(QvNode *, const char *);
    static QvNode *	readReference(QvInput *in);
    static QvBool	readNode(QvInput *in, QvName &className,QvNode *&node);
    static QvBool	readNodeInstance(QvInput *in, const QvName &className,
					 const QvName &refName, QvNode *&node);
    static QvNode *	createInstance(QvInput *in, const QvName &className);
    static QvNode *	createInstanceFromName(const QvName &className);
    static void		flushInput(QvInput *in);
};

#endif /* _QV_NODE_ */
