#include <QvDB.h>
#include <QvInput.h>
#include <QvNode.h>
#include <QvNodes.h>
#include <QvState.h>

#include <appelles.h>

class QvSeparator;

void Traverse(QvNode *n);

main(int argc, char **argv)
{
    QvDB::init();

    QvInput	in;
    QvNode	*root;

    if (QvDB::read(&in, root) && root != NULL) {
	root->ref();
	printf("Read was ok\n");
    }
    else {
	printf("Read was bad\n");
	return 1;
    }

    Geometry = Traverse(root);

    root->unref();
    
    
}

int indent = 0;

#define NODE_TYPE(n) (n)->getKind()
#define CHILD_IS_TYPE(n,i,knd) ((n)->getChild((i))->getKind() == knd##_node)
#define CHILD_TYPE(n,i) ((n)->getChild((i))->getKind())

#define MATRIX(n)       
#define TRANSFORM(n)    
#define MATERIAL(n)     
#define TRANSLATE(n)    Apply(Translate(n->translation.value[0], \
					n->translation.value[1], \
					n->translation.value[2])  
#define ROTATE(n)       Apply(Rotate(n->rotation.angle,    \
				     n->rotation.axis[0],  \
				     n->rotation.axis[1],  \
				     n->rotation.axis[2])
#define SCALE(n)        Apply(Scale(n->scaleFactor.value[0], \
				    n->scaleFactor.value[1], \
				    n->scaleFactor.value[2])


#define CUBE(n)         Apply(Scale(n->width.value,           \
				    n->height.value,          \
				    n->depth.value),          \
			      cube)
#define CONE(n)         Apply(Scale(2*n->bottomRadius.value,  \
				    2*n->bottomRadius.value,  \
				    n->heigh.valuet),         \
			      Cone(TRUE,10))
#define CYLINDER(n)     
#define SPHERE(n)       

QvBool  needsAddition(QvGroup *node, int i)
{
    // Child node 'i' needs to be added '+' to the 
    // following nodes iff:
    //   - There exists a child after it AND
    //   - The next child returns a geometry
    //     - Any node that modifies a geometry (tranform, etc...) which will be applied to a geometry
    switch(CHILD_TYPE(node,i))
    {
	// ---------------------------------------------------
	// Geometry modification nodes: these are always applied.
	// ---------------------------------------------------
    case QvTransform_node:
    case QvRotation_node:
    case QvMatrixTransform_node:
    case QvTranslation_node:
    case QvScale_node:
	// ---------------------------------------------------
	// Some other nodes are applied (updated as RBML is defined)
	// ---------------------------------------------------
    case QvMaterial_node:
	return FALSE;
	break;
    default:
	break;
    }
    return TRUE;
}

// 'returns' the next node if any.
#define NEXT_NODE(node,nChildren,i)   \
(nChildren <= 0 ? NULL :            \
 (i+1 >= nChildren ? NULL :         \
  (node)->getChild(i+1)))          

Geometry ParseXform(QvNode *parent, int child, 
		    int childCount, int *err))
{
    QvNode *current = parent->getChild(child);
    QvNode *next = NULL;
    // If there is a next child
    if(childCount > 0 && child+1 < childCount)
    {
	next = parent->getChild(child+1);
    }

    // It is wrong to have an applicative
    // expression without a next child
    if(next == NULL) {
	*err = TRUE;
	return emptyGeom;
    }

    switch(CHILD_TYPE(parent,child))
    {
    case QvTransform_node:
	{
	    QvTransform *node = (QvTransform *)n;
	    return TRANSFORM(node);
	    break;
	}
    case QvRotation_node:
	{
	    QvRotation *node = (QvRotation *)n;
	    return ROTATE(node);
	    break;
	}
    case QvMatrixTransform_node:
	{
	    QvMatrixTransform *node = (QvMatrixTransform *)n;
	    return MATRIX(node);
	    break;
	}
    case QvTranslation_node:
	{
	    QvTranslation *node = (QvTranslation *)n;
	    return TRANSLATE(node);
	    break;
	}

    case QvScale_node:
	{
	    QvScale *node = (QvScale *)n;
	    return SCALE(node);
	    break;
	}
	
    }
	
void ParseGroupNode(QvGroup *node, QvBool doParen)
{
    int close_parens = 0;
    int numKids = node->getNumChildren();
    if(doParen) 	OPEN_PAREN;
    printf("Group");
    indent++;
    for (int i = 0; i < numKids; i++) {
	Traverse(node->getChild(i));
	if(i+1 < numKids) {              // at least 1 child remains
	    if(needsAddition(node,i)) {  // Add to it
		PLUS;
	    }
	    else {	                 // Apply to it
		if(doParen) {
		indent++;
		OPEN_PAREN;
		close_parens++; }
	    }
	}
    }
    if(doParen) {
	while(close_parens) { CLOSE_PAREN; close_parens--; indent--;}
    }
    indent--;
    if(doParen)
	CLOSE_PAREN;
}

int foo(int a[])
{
    return a[0]+a[1]+a[2];
}

Geometry Traverse(QvNode *n, int i)
{
    if(n == NULL) return;

    switch (NODE_TYPE(n))
    {
	// --------------------------------------
	// Group type nodes, they have children !
	// --------------------------------------
    case QvGroup_node:
	return 
	    ParseGroupNode((QvGroup *)n, FALSE);
	break;

    case QvSeparator_node:
	ParseGroupNode((QvGroup *)n, TRUE);
	break;

    case QvLOD_node:
    case QvSwitch_node:
    case QvTransformSeparator_node:

	// these have children too (sublclass of group node)
    case QvWWWAnchor_node:
    case QvWWWInline_node:
    case QvInfo_node:
    case QvUnknownNode_node:

	// ---------------------------------------------
	// Property Nodes.  Some of these are meaningful
	// to RBML.
	// ---------------------------------------------
    case QvCoordinate3_node:
	break;
    case QvFontStyle_node:	
	break;
    case QvMaterial_node:
	{
	    QvMaterial *node = (QvMaterial *)n;
	    MATERIAL(node);
	    break;
	}
    case QvMaterialBinding_node:
    case QvNormal_node:
    case QvNormalBinding_node:	
    case QvShapeHints_node:		
    case QvTextureCoordinate2_node:	
    case QvTexture2_node:	
    case QvTexture2Transform_node:	

	// ---------------------------------------------------
	// Typed Property Nodes.  Some of these are meaningful
	// to appelles.  some... aren't.
	// ---------------------------------------------------
    case QvDirectionalLight_node:
    case QvPointLight_node:
    case QvSpotLight_node:
	printf("Light(...)");
	break;

	// ---------------------------------------------------
	// Camera nodes.  Not sure how to implement 
	// this in RBML (in a scoped way).
	// ---------------------------------------------------
    case QvOrthographicCamera_node:
    case QvPerspectiveCamera_node:
	printf("Camera(...)");
	break;

	// ---------------------------------------------------
	// Geometry modification nodes
	// ---------------------------------------------------
    case QvTransform_node:
	{
	    QvTransform *node = (QvTransform *)n;
	    TRANSFORM(node);
	    break;
	}
    case QvRotation_node:
	{
	    QvRotation *node = (QvRotation *)n;
	    ROTATE(node);
	    break;
	}
    case QvMatrixTransform_node:
	{
	    QvMatrixTransform *node = (QvMatrixTransform *)n;
	    MATRIX(node);
	    break;
	}
    case QvTranslation_node:
	{
	    QvTranslation *node = (QvTranslation *)n;
	    TRANSLATE(node);
	    break;
	}

    case QvScale_node:
	{
	    QvScale *node = (QvScale *)n;
	    SCALE(node);
	    break;
	}
	
	// ---------------------------------------------------
	// Geometry expression nodes
	// ---------------------------------------------------
    case QvAsciiText_node:
	// NOT IMPLEMENTED.
	break;
    case QvCone_node:
	{
	    QvCone *node = (QvCone *)n;
	    CONE(node);
	    break;
	}
    case QvCube_node:
	{
	    QvCube *node = (QvCube *)n;
	    CUBE(node);
	    break;
	}
    case QvCylinder_node:
	{
	    QvCylinder *node = (QvCylinder *)n;
	    CYLINDER(node);
	    break;
	}
    case QvIndexedFaceSet_node:
    case QvIndexedLineSet_node:
	// NOT IMPLEMENTED.
	break;
    case QvPointSet_node:
    case QvSphere_node:
	{
	    QvSphere *node = (QvSphere *)n;
	    SPHERE(node);
	    break;
	}

    default:
	break;
    }
}
